/*
  Generator class is the core class of the Meeting Protocol Generator
  because its purpose is to generate HTML document of the data taken
  from SQL Server.

  Alexander Podoplelov, RISC-Linz, 1998.
*/

// ****************************************************
import java.io.*;
import java.sql.*;
import java.util.*;
// ****************************************************

public class Generator {

    public   static final int      PRINTED_VERSION     = 1;
    public   static final int      BROWSABLE_VERSION   = 2;
    public   static final int      FIELD               = 1;
    public   static final int      DATA                = 2;
    public   static final int      TIMESTAMP           = 3;
    public   static final int      WATERMARK           = 4;
    public   static final int      REFERENCES          = 5;

    private  static final String[] DocVerSuffix        = {"n", "p", "b"};
    private  static final String   MeetingDirPrefix       = "M";
    private  static final String   MeetingDirHashFile     = "Meetings.hash";
    private  static final String   GeneratedIndexFileName = "Index.html";
    private  static final String   GeneratedDataFilePrefix = "Data";
    private  static final String   GeneratedDataFileSuffix = ".html";
    private  static final String   WWW_CACHE_NAME         = "index.cache";

    private  static Vector IndexHeader      = new Vector();
    private  static Vector IndexListStart   = new Vector();
    private  static Vector IndexListItem    = new Vector();
    private  static Vector IndexListEnd     = new Vector();
    private  static Vector IndexFooter      = new Vector();

    private  static Vector DataHeader       = new Vector();
    private  static Vector DataListStart    = new Vector();
    private  static Vector DataListItem     = new Vector();
    private  static Vector DataListEnd      = new Vector();
    private  static Vector DataFooter       = new Vector();

    private  static Vector BannerHeader     = new Vector();
    private  static Vector BannerListStart  = new Vector();
    private  static Vector BannerListItem   = new Vector();
    private  static Vector BannerListEnd    = new Vector();
    private  static Vector BannerFooter     = new Vector();
    private  static Hashtable CurrentGenMeeting  = new Hashtable(); 

    public   static Properties MeetingHash  = new Properties(); //synchronized

    private  ShelledConnection     connection;
    private  MeetingItem           current_meeting;
    private  BufferedWriter        www_cache;
    
    public static void Init() {
	ReadDocTemplate
	    (Configuration.IndexTemplate,
	     IndexHeader,
	     IndexListStart, IndexListItem, IndexListEnd,
	     IndexFooter
	     );
	ReadDocTemplate
	    (Configuration.DataTemplate,
	     DataHeader,
	     DataListStart, DataListItem, DataListEnd,
	     DataFooter
	     );
	ReadDocTemplate
	    (Configuration.BannerTemplate,
	     BannerHeader,
	     BannerListStart, BannerListItem, BannerListEnd,
	     BannerFooter
	     );

	LoadMeetingHash();
    }

    public Generator (ShelledConnection connection, MeetingItem mi) {
	this.connection       = connection;
	this.current_meeting  = mi;
    }

    private void AquareMeetingLock (int meeting_code) {
	Lock lock;
	Integer key = new Integer (meeting_code);
	synchronized (CurrentGenMeeting) {
	    lock = (Lock)CurrentGenMeeting.get (key);
	}
	if (lock == null)
	    synchronized (CurrentGenMeeting) {
		CurrentGenMeeting.put (key, new Lock (true));
	    }
	else {
	    while (lock.locked());
	    synchronized (lock) {
		lock.lock();
	    }
	}
    }

    private void ReleaseMeetingLock (int meeting_code) {
	Lock lock;
	Integer key = new Integer (meeting_code);
	synchronized (CurrentGenMeeting) {
	    lock = (Lock)CurrentGenMeeting.get (key);
	}
	if (lock != null)
	    synchronized (lock) {
		lock.unlock();
	    }
    }

    public String GenerateDoc (int doc_type)
	throws ErrorException
    {
	String res_url     = null;
	String scode       = getMeetingDirID (current_meeting.nn, doc_type);
	String stime       = null;
	File   mdir        = new File (Configuration.HardDrive_base, scode);
	
	synchronized (MeetingHash) {
	    stime = MeetingHash.getProperty (scode);
	}
	try {
	    if (stime == null || !mdir.exists()) {
		AquareMeetingLock (current_meeting.nn);
		CreateDocument (mdir, doc_type);
		ReleaseMeetingLock (current_meeting.nn);
	    }
	} catch (DocProtException e) {
	    MeetingProtocolGenerator.Log (e.toString());
	    return Configuration.WWW_base + Configuration.ErrorHTML;
	}

	res_url = Configuration.WWW_base + scode + "/" +
	    GeneratedIndexFileName;

	synchronized (MeetingHash) {
	    MeetingHash.put
		(scode, Long.toString (new java.util.Date().getTime()));
	}
	SaveMeetingHash();
	
	return res_url;
    }

    public static synchronized void LoadMeetingHash() {
	try {
	    FileInputStream f = new FileInputStream (MeetingDirHashFile);
	    MeetingHash.load (f);
	    f.close();
	}
	catch (FileNotFoundException e) {}
	catch (IOException e1) { Configuration.abort (e1.toString()); }
    }

    public static synchronized void SaveMeetingHash() {
	try {
	    File f = new File (MeetingDirHashFile);
	    if (f.exists())
		f.delete();
	    FileOutputStream fo = new FileOutputStream (f);
	    MeetingHash.save
		(fo,
		 "Hash table for meetings. A. Podoplelov, 1998, RISC-Linz."
		 );
	    fo.close();
	}
	catch (IOException e) { Configuration.abort (e.toString()); }
    }

    private void CreateDocument (File dir, int d_t) throws DocProtException {
	Vector contributions = LoadFromDB (current_meeting.getID(), d_t);
	File f = new File (dir.getPath(), GeneratedIndexFileName);
	try {
	    dir.mkdir();
	    www_cache = new BufferedWriter
		(new FileWriter (new File (dir.getPath(), WWW_CACHE_NAME)));
	    www_cache.write ("serveall=true\n\n");
	    www_cache.write ("file=" + GeneratedIndexFileName +
			     "&content=text/html&title=Title\n");
	    BufferedWriter bw = new BufferedWriter (new FileWriter (f));
	    MakeIndex (bw, contributions,
		       IndexHeader,
		       IndexListStart, IndexListItem, IndexListEnd,
		       IndexFooter);
	    bw.close();
	    MakeData (dir, contributions, d_t);
	    bw = new BufferedWriter
		(new FileWriter (new File (dir.getPath(),
					   Configuration.DefDataFile)));
	    www_cache.write ("file=" + Configuration.DefDataFile +
			     "&content=text/html&title=Title\n");
	    MakeIndex (bw, contributions,
		       BannerHeader,
		       BannerListStart, BannerListItem, BannerListEnd,
		       BannerFooter);
	    bw.close();
	    www_cache.close();
	    www_cache = null;
	} catch (IOException e) {
	    MeetingProtocolGenerator.Log (e.toString());
	}
    }

    // Reads data of Contribution Items and corresponding Contribution
    // Parts. It also re-arranges Contribution parts according to
    // "parent-child" structure.
    private Vector LoadFromDB (int meeting, int d_t) throws DocProtException {
	Vector contr = null;
	try {
	    contr = connection.GetContributions (meeting);
	    if (contr == null)
		throw new DocProtException ("no records");
	    for (int i = 0; i < contr.size(); i++) {
		ContributionItem ci = (ContributionItem)contr.elementAt (i);
		ci.setContributionParts
		    (connection.GetContributionParts (ci.getID()));
		ci.setReferences
		    (connection.GetReferences (ci.getID()));
		if (ci.ContributionParts() != null )
		    switch (d_t) {
		    case PRINTED_VERSION:
			ci.setWaterMark
			    (GeneratedDataFilePrefix +
			     GeneratedDataFileSuffix + "#" + ci.getID());
			break;
		    case BROWSABLE_VERSION:
			ci.setWaterMark
			    (GeneratedDataFilePrefix + ci.getID() +
			     GeneratedDataFileSuffix);
			break;
		    }
		else
		    ci.setWaterMark (Configuration.WWW_base +
				     Configuration.NoPartsHTML);
	    }
	} catch (SQLException e) {
	    throw new DocProtException (e.toString());
	}

	// Completing reference records
	for (int i = 0; i < contr.size(); i++) {
	    ContributionItem ci = (ContributionItem)contr.elementAt (i);
	    if (ci.References() != null)
		for (int j = 0; j < ci.References().size();) {
		    int k;
		    Reference r = (Reference)ci.References().elementAt (j);
		    if (r.to_ci == null) {
			for (k = 0; k < contr.size(); k++) { // Linear search
			    ContributionItem ci1 =
				(ContributionItem)contr.elementAt (k);
			    if (ci1.getID() == r.to) {
				r.to_ci = ci1;
				break;
			    }
			}
			if (k < contr.size()) // We found the corresp. record
			    j++;
			else // No corresponding Contribution, delete reference
			    ci.References().removeElementAt (j);
		    } else
			j++;
		}
	}

	// Re-arranging according to tree structure of "parent" fields
	Vector result = new Vector();
	for (int j = 0; j < contr.size(); j++) {
	    ContributionItem ci = (ContributionItem)contr.elementAt (j);
	    for (int  i = j; i < contr.size(); i++) {
		ContributionItem ci1 = (ContributionItem)contr.elementAt (i);
		if (ci1.Parent() == ci.ID()) {
		    if (ci.Childs() == null)
			ci.NewChilds();
		    ci.Childs().addElement (ci1);
		} 
	    }
	    if (ci.Parent() == 1)
		result.addElement (ci);
	}
	contr = null; // for garbage collector, let him eat it.
	return result;
    }

    private String getMeetingDirID (int meeting_id, int doc_type) {
	return
	    MeetingDirPrefix +
	    Integer.toString(meeting_id) +
	    DocVerSuffix [doc_type];
    }
    
    // Creates index file for Contribution Items (root proc.)
    private void MakeIndex (BufferedWriter f, Vector c,
			    Vector Header,
			    Vector ListStart,
			    Vector ListItem,
			    Vector ListEnd,
			    Vector Footer) throws IOException {
	int i;
 	for (i = 0; i < Header.size(); i++)
 	    ((DocPortion)Header.elementAt (i)).WriteValue
 		(f, current_meeting, null, null, false);

	IndexList (f, c, Header, ListStart, ListItem, ListEnd, Footer);
	
 	for (i = 0; i < Footer.size(); i++)
 	    ((DocPortion)Footer.elementAt (i)).WriteValue
 		(f, current_meeting, null, null, false);

	f.flush();
    }

    // Creates index list (recursevely) file for Contriubution Items
    private void IndexList (BufferedWriter f, Vector c,
			    Vector Header,
			    Vector ListStart,
			    Vector ListItem,
			    Vector ListEnd,
			    Vector Footer) throws IOException {
	int i, j, k;

 	for (i = 0; i < ListStart.size(); i++)
	    ((DocPortion)ListStart.elementAt (i)).WriteValue
 		(f, current_meeting, null, null, false);
	
 	for (j = 0; j < c.size(); j++) {
 	    ContributionItem ci = (ContributionItem)c.elementAt (j);
 	    for (i = 0; i < ListItem.size(); i++) 
 		((DocPortion)ListItem.elementAt (i)).WriteValue
 		    (f, current_meeting, ci, null, false);
	    if (ci.Childs() != null)
		    IndexList (f, ci.Childs(), 
			       Header, ListStart, ListItem, ListEnd, Footer);
 	}

 	for (i = 0; i < ListEnd.size(); i++)
 	    ((DocPortion)ListEnd.elementAt (i)).WriteValue
 		(f, current_meeting, null, null, false);
    }

    // Makes data set of Contribution Part Items files or one file for 
    // printed version
    private void MakeData (File dir, Vector c, int doc_type)
	throws IOException
    {
	File f;
	BufferedWriter bw;
	
	switch (doc_type) {
	case PRINTED_VERSION:
	    f = new File (dir.getPath(), GeneratedDataFilePrefix +
			  GeneratedDataFileSuffix);
	    www_cache.write ("file=" + GeneratedDataFilePrefix +
			     GeneratedDataFileSuffix +
			     "&content=text/html&title=Title\n");
	    bw = new BufferedWriter (new FileWriter (f));
	    PrintData (bw, c, true);
	    bw.close();
	    break;
	case BROWSABLE_VERSION:
	    PrintData (dir, c);
	    break;
	}
    }

    // Prints Contribution Parts of each Contribution Item
    // in the vector c and then looks into its childs recursively.
    // The result is in one file.
    private void PrintData (BufferedWriter f, Vector c, boolean root)
	throws IOException
    {
	int i;
	for (i = 0; i < c.size(); i++) {
	    ContributionItem ci = (ContributionItem)c.elementAt (i);
	    if (ci.ContributionParts() != null)
		MakeDataPortion (f, ci, false);
	    if (ci.Childs() != null)
		PrintData (f, ci.Childs(), false);
	}
	if (root) 
	    for (i = 0; i < DataFooter.size(); i++)
		((DocPortion)DataFooter.elementAt (i)).WriteValue
		    (f, current_meeting, null, null, true);
    }


    // Separately by files, prints Contribution Parts of each Contribution Item
    // in the vector c and then looks into its childs recursively
    // 
    private void PrintData (File dir, Vector c)
	throws IOException
    {
	for (int i = 0; i < c.size(); i++) {
	    ContributionItem ci = (ContributionItem)c.elementAt (i);
	    if (ci.ContributionParts() != null) {
		File f = new File (dir.getPath(), GeneratedDataFilePrefix +
			      ci.getID() + GeneratedDataFileSuffix);
		www_cache.write ("file=" + GeneratedDataFilePrefix +
				 ci.getID() + GeneratedDataFileSuffix +
				 "&content=text/html&title=Title\n");
		BufferedWriter bw = new BufferedWriter (new FileWriter (f));
		MakeDataPortion (bw, ci, true);
		bw.close();
	    }
	    if (ci.Childs() != null)
		PrintData (dir, ci.Childs());
	}
    }

    // Prints the sequence of Contribution Part Items into the stream
    // for particular Contribution Item
    private void MakeDataPortion (BufferedWriter f,
				  ContributionItem c,
				  boolean print_footer)
	throws IOException
    {
	int i, j;
	for (i = 0; i < DataHeader.size(); i++)
	    ((DocPortion)DataHeader.elementAt (i)).WriteValue
		(f, current_meeting, c, null, true);

 	for (i = 0; i < DataListStart.size(); i++)
	    ((DocPortion)DataListStart.elementAt (i)).WriteValue
 		(f, current_meeting, c, null, true);

	for (j = 0; j < c.ContributionParts().size(); j++) {
	    ContributionPartItem cpi =
		(ContributionPartItem)c.ContributionParts().elementAt (j);
	    for (i = 0; i < DataListItem.size(); i++) 
		((DocPortion)DataListItem.elementAt (i)).WriteValue
		    (f, current_meeting, c, cpi, true);
	}
	
 	for (i = 0; i < DataListEnd.size(); i++)
 	    ((DocPortion)DataListEnd.elementAt (i)).WriteValue
 		(f, current_meeting, c, null, true);
	
	if (print_footer)
	    for (i = 0; i < DataFooter.size(); i++)
		((DocPortion)DataFooter.elementAt (i)).WriteValue
		    (f, current_meeting, c, null, true);

	f.flush();
    }

    // Reads template for creating result HTML or whatever document.
    private static void ReadDocTemplate (String fname, 
					 Vector header,
					 Vector body_list_start,
					 Vector body_list_item,
					 Vector body_list_end,
					 Vector footer)
    {
	FileReader         f   = null;
	DocumentProtocol   dp  = null;
	try {
	    f  = new FileReader (fname);
	    dp = new DocumentProtocol (f);
	} catch (FileNotFoundException e) {
	// The existance of the file we have already 
	// checked before in the "Configuration.Check()" procedure.
	}
	try {
	    while (true) {
		DocMessage message = dp.GetMessage();
		switch (message.key) {
		case Message.DATA:
		    // Skip this data because it does not belong
		    // to any command
		    break;
		case Message.CMD_HEADER:
		    ComposeVector (header, dp);
		    break;
		case Message.CMD_FOOTER:
		    ComposeVector (footer, dp);
		    break;
		case Message.CMD_ITERATE:
		    boolean done = false;
		    while (!done) {
			message = dp.GetMessage();
			switch (message.key) {
			case Message.DATA:
			    // Skip this data because it does not belong
			    // to any command within ITERATE
			    break;
			case Message.CMD_END:
			    done = true;
			    break;
			case Message.CMD_LIST_START:
			    ComposeVector (body_list_start, dp);
			    break;
			case Message.CMD_LIST_END:
			    ComposeVector (body_list_end, dp);
			    break;
			case Message.CMD_LIST_ITEM:
			    ComposeVector (body_list_item, dp);
			    break;
			default:
			    throw new DocProtException ("no such key: " +
							message.key);
			}
		    }
		    break;
		default:
		    throw new DocProtException ("no such key: " + message.key);
		} //switch
	    } // while
	} catch (SessionDoneException e) {
	} catch (DocProtException e1) {
	    Configuration.abort ("Generator: " + e1.toString());
	}
	try { f.close(); } catch (IOException e) {
	    Configuration.abort ("Generator: " + e.toString());
	}
    }

    // Reads components for a particular vector among Header, Footer and 
    // Lister ones
    private static void ComposeVector (Vector collector, DocumentProtocol dp)
	throws DocProtException, SessionDoneException
    {
	while (true) {
	    DocMessage dm = dp.GetMessage();
	    if (dm.key == Message.CMD_END) return;
	    if (dm.key == Message.CMD_FIELD) {
		dm = dp.GetMessage();
		if (dm.key != Message.DATA)
		    throw new DocProtException ("Syntax error: COMMAND");
		collector.addElement
		    (new DocPortion (FIELD, (String)dm.data.elementAt (0)));
	    } else if (dm.key == Message.TIMESTAMP)
		collector.addElement (new DocPortion (TIMESTAMP, "*:**"));
	    else if (dm.key == Message.DATA)
		collector.addElement
		    (new DocPortion (DATA, (String)dm.data.elementAt (0)));
	    else if (dm.key == Message.WATERMARK)
		collector.addElement (new DocPortion (WATERMARK, null));
	    else if (dm.key == Message.CMD_REFERENCES)
		collector.addElement (new DocPortion (REFERENCES, null));
	}
    }
}

// ****************************************************
class Lock extends Object {
    private boolean is_locked;

    public Lock (boolean is_locked) {
	super();
	this.is_locked = is_locked;
    }

    public void lock() {
	is_locked = true;
    }

    public void unlock() {
	is_locked = false;
    }

    public boolean locked () {
	return is_locked;
    }
}

// ****************************************************
class DocPortion extends Object {
    protected int     type;
    protected int     table;
    protected String  info;

    public DocPortion (int type, String Data) throws DocProtException {
	super();
	this.type  = type;
	this.table = 0;
	switch (type) {
	case Generator.FIELD:
	    int separator = Data.indexOf ((int)'.');
	    if (separator < 0)
		throw new DocProtException
		    ("No table- or field-name in \"" + Data + "\".");
	    this.table =
		MeetingSystemDB.getTableCode (Data.substring (0, separator));
	    this.info = Data.substring (separator + 1);
	    break;
	case Generator.DATA:
	case Generator.REFERENCES:
	case Generator.TIMESTAMP:
	    this.info = Data;
	    break;
	}
    }

    public String getInfo() {
	switch (type) {
	case Generator.FIELD:
	    return "(" + table + ", " + info + ")";
	default:
	    //	case Generator.DATA:
	    //	case Generator.TIMESTAMP:
	    return info;
	}
    }

    public void WriteValue (BufferedWriter f, MeetingItem mi,
			    ContributionItem ci, ContributionPartItem cpi,
			    boolean watermarkshortcut)
	throws IOException
    {
	switch (type) {
	case Generator.DATA:
	    f.write (info);
	    break;
	case Generator.WATERMARK:
	    f.write (ci.getWaterMark(watermarkshortcut));
	    break;
	case Generator.TIMESTAMP:
	    f.write (Configuration.Format (new java.util.Date()));
	    break;
	case Generator.REFERENCES:
	    ci.WriteReferences (f);
	    break;
	case Generator.FIELD:
	    switch (table) {
	    case 1: mi.Write (f, info); break; // meeting
	    case 4: ci.Write (f, info); break; // contribution
	    case 5: cpi.Write (f, info); break; // contribution_part
	    }
	    break;
	}
    }
}
// ****************************************************

