/*
 Meeting Protocol Generator System.
 Goal: Accepting requests from the network by TCP/IP protocol
       and generating a bunch of HTML files reflecting information
       about selected Meeting. The files are placed into WWW visible
       area so that it is possible to acces them though http protocol.
       Thus, an URL to these files is sent back to the caller.
       Database is accessed by JDBC using Openlink Client classes.

 A. Podoplelov, RISC-Linz, 1998.
*/

// ----------------------------------------------------------------
import java.net.*;
import java.io.*;
import java.util.*;
import java.util.Date;
import java.sql.*;
import java.lang.*;
import java.text.*;
// ----------------------------------------------------------------

public class MeetingProtocolGenerator implements Runnable {
    // ---------------------------------------------------------------
    // class constants
    // ---------------------------------------------------------------
    public  static final int          MAX_DB_CONNECTIONS = 3;
    public  static final String       SIGNATURE = 
      "Research in Symbolic Computation Institute (RISC), 1998.";
    public  static final String       VERSION = 
      "Meeting Protocol Generator System V 1.0";
    public  static final int          PORT = 7239;
    protected static final boolean    SIMULATION = true;

    // ---------------------------------------------------------------
    // class variables
    // ---------------------------------------------------------------
    //    public  static Vector       accepted_clients = null;
    protected static SecurityManager  securitymanager = null;

    // ---------------------------------------------------------------
    // object variables
    // ---------------------------------------------------------------
    protected            Vector                 accepted_clients = null;
    private              boolean                debug   = false;
    private              int                    port    = PORT;
    private              Vulture                vulture = null;
    private              DirectoryWatchDog      eater   = null;
    // ------------------
    // Main body
    // ------------------
    public static void main (String[] arg) {
	//	mpDataFormat = DateFormat.getDateInstance();
	//	mpCalendar   = Calendar.getInstance();
	SecurityManager securitymanager = System.getSecurityManager();
	if (securitymanager == null)
	    System.out.println ("Warning: No Security Manager.");
	System.out.println (SIGNATURE + "\n" + VERSION);
	Configuration.Check();
	Generator.Init();
	new MeetingProtocolGenerator(arg).run();
	System.exit (0);
    }

    public MeetingProtocolGenerator(String[] arg) {
	Init();
	ParseArgs (arg);
    }

    // Starting the server
    public void run () {
	StartDaemon();
    }

    private void Init() {
	accepted_clients = new Vector();
	if (!debug) {
	    vulture = new Vulture (this);
	    eater = new DirectoryWatchDog
		(new File(Configuration.HardDrive_base));
	}
	debug = false;
	port = PORT;
    }

    private void StartDaemon() {
	if (debug) {
	    Log ("Debug mode is activated.");
	    BufferedWriter out =
		new BufferedWriter (new OutputStreamWriter(System.out));
	    BufferedReader in = 
		new BufferedReader (new InputStreamReader(System.in));

	    new ClientSession (in, out).run();
	    Log ("Meeting Protocol Generator is Done.");
	} else
	    try {
		AcceptClients();
	    } catch (IOException e) {
		e.printStackTrace();
		abort ("IOException while accepting clients.");
	    }
    }

    // ----------------------------------------------------------------------
    // Main loop of the Server. It accepts clients and creates the
    // corresponding serving object for each of them.
    // ----------------------------------------------------------------------
    private void AcceptClients () throws IOException {
	Log ("Capturing TCP/IP port number " + Integer.toString(port) + ".");
	ServerSocket ss = new ServerSocket (port);
	Socket guest = null;

	while (true) {
	    Log ("the server is ready");
	    try {
		guest = ss.accept();
		int id  = 
		    ThreadClientSession (new ClientSession (guest, vulture));
		Log("Accepting \"" + guest.getInetAddress().getHostName()
		    + "\". ID=" + id + ".");
	    } catch (IOException e) {
		Log ("Error in accepting client socket. " +
		     "Discrading this connection." + e.toString());
	    }
	}
    }

    // ----------------------------------------------------------------------
    // Collects all the client threads into a vector 
    // ----------------------------------------------------------------------
    private int ThreadClientSession (ClientSession client) {
	synchronized (accepted_clients) {
	    accepted_clients.addElement (client);
	    client.start();
	}
	return client.id;
    }

    private void ParseArgs (String[] args) {

	// parse arguments
	if (args.length < 1)
	    return;

	String key = args[0];
	Log ("Argument : \"" + key + "\"");
	if (key.equals("-debug"))
	    debug = true;
	else if (key.equals("-help")) {
	    printHelp();
	    System.exit(0);
	} else if ( key.length() > 6 && key.substring(0, 5).equals("-port")) {
	    try {
		port = Integer.parseInt (key.substring (6, key.length()));
	    } catch (NumberFormatException e) {
		argError();
	    }
	} else
	    argError();
    }
    
    // ----------------------------------------------------------------------
    // abort operation because of error in arguments
    // ----------------------------------------------------------------------
    private static void printHelp() {
	System.out.println("arguments: '-port=<port number>' start the server with specific TCP/IP port");
	System.out.println("         | '-debug' standard input/output");
	System.out.println("         | '-help'  this help");
    }

    // ----------------------------------------------------------------------
    // abort operation because of error in arguments
    // ----------------------------------------------------------------------
    private static void argError() {
	System.err.println("Wrong argument.");
	printHelp();
	System.exit (-1);
    }

    public static String Date() {
	Date d = new Date();
	return d.toString().substring (4, 19);
    }

    // ----------------------------------------------------------------------
    // printing log information
    // ----------------------------------------------------------------------
    protected static synchronized void Log (String info) {
	System.out.println ("[" + Date() + "] " + info);
    }

    // ----------------------------------------------------------------------
    // abnormal termination
    // ----------------------------------------------------------------------
    private static void abort (String info) {
	System.err.println (Date() + " " + info);
	System.exit (-1);
    }
}

// --------------------------------------------------------------------------
// Client connection serving object
// --------------------------------------------------------------------------
class ClientSession extends Thread {
    public    static final int   IDLIMIT = 100;
    protected static int         counter = 0;
    protected int                id;
    private   Vulture            vulture;
    protected Socket             source;
    private   ShelledConnection  connection;
    private   CommunicationProtocol protocol;
    private   Vector             meetings         = null;

    public ClientSession (Socket source, Vulture vulture) throws IOException {
	protocol = new ServerCommunicationProtocol
	    (new BufferedReader
	     (new InputStreamReader(source.getInputStream())),
	     new BufferedWriter
	     (new OutputStreamWriter(source.getOutputStream()))	    
	     );
	this.source  = source;
	this.vulture = vulture;
	connection   = MeetingSystemDB.RequestConnection();
	id = ++counter;
    }

    public ClientSession (BufferedReader in, BufferedWriter out) {
	protocol = new ServerCommunicationProtocol (in, out);
	this.source  = null;
	this.vulture = null;
	connection   = MeetingSystemDB.RequestConnection();
	id = ++counter;
    }

    public void run () {
	try {
	    Schedule();
	} catch (SessionDoneException e) {
	    if (source != null)
		try { source.close(); } catch (IOException e1) {}
	    if (vulture != null)
		synchronized (vulture) {
		    MeetingProtocolGenerator.Log ("Notifying vulture object.");
		    vulture.notify();
		}
	    MeetingProtocolGenerator.Log
		("ClientSession is done. ID=" + id + ".");
	}
    }

    // This scheduler is pretty straightforward 
    private void Schedule() throws SessionDoneException {
	try {
	    while (true)
		try {
		    Dispatch (protocol.GetMessage());
		} catch (UnknownKeyException e) {
		    MeetingProtocolGenerator.Log (e.toString());
		    e.printStackTrace();
		}
	} catch (IOException e1) {
	    e1.printStackTrace();
	    throw new SessionDoneException(e1.toString());
	}
    }


    // ----------------------------------------------------------------------
    // Dispatching messages
    // ----------------------------------------------------------------------
    private void Dispatch (Message message) throws SessionDoneException {
	protocol.SendMessage (Message.CONFIRMATION, message.key);
	try {
	    switch (message.key) {
		// Sending signature information
	    case Message.HANDSHAKE:
		protocol.SendMessage (Message.HANDSHAKE,
				      MeetingProtocolGenerator.VERSION);
		break;
		// Sending list of available meetings
	    case Message.MEETINGLIST:
		if (meetings == null) {
		    //		    if (!MeetingProtocolGenerator.SIMULATION)
		    meetings = connection.GetMeetings();
		    //		    else
		    // meetings = MeetingSystemDB.Meetings;
	        }
		protocol.SendMessage (Message.MEETINGLIST, meetings);
		break;
		// Sending list of available meetings
	    case Message.TOOLS:
		protocol.SendMessage (Message.TOOLS, connection.GetTools());
		break;
		// Generating the Document
	    case Message.GENERATE:
		Generator generator = new Generator
		    (connection,
		     (MeetingItem)meetings.elementAt (Integer.parseInt
		     ((String)message.data.elementAt (0)))
		     );
		String    docURL    = generator.GenerateDoc
		    // Doc Type
		    (Integer.parseInt ((String)message.data.elementAt (1)));
		protocol.SendMessage (Message.DOCURL, docURL);
		break;
		// Closing the connection
	    case Message.CLOSE:
	    default:
		throw new SessionDoneException
		    ("Client Session is Done.");
	    }
	} catch (SQLException e) {
	    MeetingProtocolGenerator.Log (e.toString());
	} catch (ErrorException e1) {
	    MeetingProtocolGenerator.Log (e1.toString());
	}

    }

}

// ------------------------------------------------------
// Vulture object. It cleans the vector of clients
// from "dead" threads - threads which closed the connection
// to the client. After initialization it starts automatically.

class Vulture extends Thread {
    public  static final int   VULTURE_DELAY   = 5000; // 5 seconds
    private MeetingProtocolGenerator mp        = null;
    private boolean                  reset_ids = false;

    public Vulture (MeetingProtocolGenerator mp) {
	this.mp = mp;
	this.reset_ids = false;
	this.start();
    }

    // It's syncrinized in order to call 'this.wait()'
    public synchronized void run () {
	MeetingProtocolGenerator.Log
	    ("Thread \"eater\" for \"dead\" client sessions has been started");
	while (true) {
	    try {
		this.wait (VULTURE_DELAY);
	    } catch (InterruptedException e) {
		MeetingProtocolGenerator.Log ("Vulture got notification.");
	    }

	    // prevent simultaneous access
	    synchronized (mp.accepted_clients) {
		for (int i = 0; i < mp.accepted_clients.size(); i++) {
		    ClientSession cs =
			(ClientSession)mp.accepted_clients.elementAt (i);
		    if (!cs.isAlive()) {
			int id = cs.id;
			mp.Log ("Removing thread of \"" +
			     cs.source.getInetAddress().getHostName() +
			     "\", ID=" + id + ".");
			mp.accepted_clients.removeElementAt (i);
			if (!reset_ids && id > ClientSession.IDLIMIT)
			    reset_ids = true;
			if (reset_ids && mp.accepted_clients.size() == 0) {
			    mp.Log("Reseting ClientSession IDs.");
			    ClientSession.counter = 0;
			    reset_ids = false;
			}
		    }
		} // for
	    } // synchronized
	} // while
    }
}
// ------------------------------------------------------
